import pandas as pd
import cPickle as pickle
import math
import os
import sys

def calculate_posCount(region_aaCount):
	"""Function that calculates the total number of amino acids at positions in antibody regions
	Inputs:	
		@region_aaCount: Counts of amino acids at positions in a given antibody region (e.g. framework 1)
	Return:
		@pos_Count: The total number of amino acids at positions
	"""
	pos_Count = {}

	for pos in region_aaCount:
		for ele in region_aaCount[pos]:
			if pos not in pos_Count:
				pos_Count[pos] = 0
			pos_Count[pos] += region_aaCount[pos][ele]

	return pos_Count

def noise_removal(error,IMGT_refence,pos_amino_acids, pos_Count, framework):
	"""Function that creates a reference matrix for an antibody region
	Inputs:
		@error:     Residue error rate
		@pos_amino: Amino acid counts by position in an antibody region
		@pos_Count: Total number of amino acids at a position

	Return:
		@amino_acid_by_pos: Matrix of counts of amino acids at positions if their proportions are above the residue error rate

	"""
	bad_aminos = set("BJXZ*")
	amino_acid_by_pos = {}
	for position in pos_amino_acids:
		if position not in amino_acid_by_pos:
			amino_acid_by_pos[position] = {}
		for amino in pos_amino_acids[position]:
			if amino in bad_aminos:
				continue
			else:
				amino_pos_proportion = float(pos_amino_acids[position][amino]) / pos_Count[position]

				if position in IMGT_refence[framework]:
					if IMGT_refence[framework][position][amino] == 0:
						if amino_pos_proportion <= error:
							amino_acid_by_pos[position][amino] = 0
						else:
							amino_acid_by_pos[position][amino] = pos_amino_acids[position][amino]
					else:
						amino_acid_by_pos[position][amino] = pos_amino_acids[position][amino]
				else:
					if amino_pos_proportion <= error:
						 amino_acid_by_pos[position][amino] = 0
					else:
						amino_acid_by_pos[position][amino] = pos_amino_acids[position][amino]
	return amino_acid_by_pos

def CalculateResidueErrorRate(regions_aaCount, chain):
	"""Function calculate the residue error rate for Ig-seq Data
	Inputs:
		@regions_aaCount: Counts of amino acids at antibody positions in Ig-seq data
		@chain: Igseq data antibody chain
	Return:
		@error: The residue error rate value
	"""

	error = 0
	if chain == "H":
		error_fw1 = sorted(regions_aaCount["fwh1"]["23"].values(), key=lambda x: x, reverse=True)
		error_fw1 = [float(x) / sum(error_fw1) for x in error_fw1]
		error_fw3 = sorted(regions_aaCount["fwh3"]["104"].values(), key=lambda x: x, reverse=True)	
		error_fw3 = [float(x) / sum(error_fw3) for x in error_fw3]
	elif chain == "L":
		error_fw1 = sorted(regions_aaCount["fwl1"]["23"].values(), key=lambda x: x, reverse=True)
		error_fw1 = [float(x) / sum(error_fw1) for x in error_fw1]
		error_fw3 = sorted(regions_aaCount["fwl3"]["104"].values(), key=lambda x: x, reverse=True)	
		error_fw3 = [float(x) / sum(error_fw3) for x in error_fw3]
	if len(error_fw1) == 1 and len(error_fw3) == 1 :
		error = 0
	else:
		if len(error_fw1) > 1 and len(error_fw3) > 1:		
			if error_fw1[1] >= error_fw3[1]:
				error = error_fw1[1]
			else:
				error = error_fw3[1]
		elif len(error_fw1) == 1:
			error = error_fw3[1]
		else:
			error = error_fw1[1]
	return error
			
def frame_canon(cols):
	"""Reordering canonical CDRs and frameworks according to IMGT definition
	"""
	
	integer_list =  [int(x) for x in cols]
	sorted_integer_list = sorted(integer_list,key=int)
	sorted_string_list = [str(x) for x in sorted_integer_list]
	return sorted_string_list

def cdrh3(cols):
	"""Reordering CDRH3 or CDRL3 according to IMGT definition
	"""
	integer_list = []
	pos_111 = []
	pos_112 = []
	sorted_string_list = []
	for pos in cols:
		try:
			integer_list.append(int(pos))
			sorted_integer_list = sorted(integer_list,key=int)
			sorted_string_list = [str(x) for x in sorted_integer_list]
		except:
			if "111" in pos:
				pos_111.append(pos)
			elif "112" in pos:
				pos_112.append(pos)
	if "112A" in pos_112:
		pos_111_112_list = sorted(pos_111) + sorted(pos_112)[::-1]
		sorted_string_list = sorted_string_list[:sorted_string_list.index("111")+1] + pos_111_112_list + sorted_string_list[sorted_string_list.index("111")+1:]
	return sorted_string_list					

def all_aa_proportions(chain,frame,regions_aaCount,IMGT_refence):
	"""Function that creates a reference matrix for Ig-seq data.
	Inputs:
		@chain = Igseq data antibody chain
		@frame = antibody region (can be framework or cdr)
		@regions_aaCount = Counts of amino acids at antibody positions in Ig-seq data
	"""
	residueErrorRate = CalculateResidueErrorRate(regions_aaCount,chain)


	if os.path.isdir("Reference_matrix") != True:
		os.makedirs("Reference_matrix")
	if os.path.isdir("Error") != True:
		os.makedirs("Error")

	pos_Count = calculate_posCount(regions_aaCount[frame])

	counting_aa = noise_removal(residueErrorRate,IMGT_refence, regions_aaCount[frame], pos_Count, frame) 

	pos_Count = calculate_posCount(counting_aa)  # Here we recalculate number of entries per position after the removal of residue/positions found below the residue error rate

	####################################################
	## Turning Amino Acid counts into proportions ######
	####################################################	

	for pos in counting_aa:
		entropy = 0
		for ele in counting_aa[pos]:
			 counting_aa[pos][ele] = float(counting_aa[pos][ele]) / pos_Count[pos]
			 
	df_aa_proportion = pd.DataFrame(counting_aa).fillna(0) # Amino acid Distribution Dataframe
	pos_hist = pd.DataFrame(pos_Count, index = [0]).fillna(0) # This DataFrame is used to find positions with less than 20 entries

	###############################################
	##### Here we put positions in IMGT order #####
	###############################################
	cols = pos_hist.columns.tolist() # Column names
	for col in cols:
		if pos_hist[col][0] <= 20:  # Threshold for minimum entries at positions
			del df_aa_proportion[col]

			print "Position with fewer than 20 entries: ", col
	cols = df_aa_proportion.columns.tolist()

	if frame not in ["cdrh3","cdrl3"]:
		new_cols = frame_canon(cols)
	else:
		new_cols = cdrh3(cols)		

	##########################
	## AA proportion graph ###
	##########################		
	df_aa_proportion = df_aa_proportion[new_cols]
	df_aa_proportion = df_aa_proportion.transpose()

	####################################
	## Save the reference matrix here ##
	####################################
	with open("Reference_matrix/reference_matrix_%s.pkl" % (frame), "wb") as y:
		pickle.dump(df_aa_proportion, y)
	with open("Error/ig_seq_residue_error_rate.txt", "w") as fi:
		fi.write("\n")
		fi.write("The residues error rate for this Ig-seq dataset is %f \n" % (residueErrorRate))
	return None


if __name__ == '__main__':

	pass
